@@ -7,19 +7,16 @@ use Scalar::Util qw(blessed);
 use Carp qw(carp confess);
 use POE;
 
-use vars qw($VERSION @ISA);
-$VERSION = '0.31';
+use vars qw($VERSION);
+$VERSION = '0.41';
 sub DOSENDBACK () { 1 }
 
-our $globalinfos;
-my $debug = 0;
-
-my $PATCH         = 18;
-my $HANDSHAKE     = 19;
-my $EVENT_FLUSHED = 20;
-my $EVENT_INPUT   = 21;
+our $globalinfos = {};
 
 BEGIN {
+   our $HANDSHAKE     = 19;
+   our $EVENT_FLUSHED = 20;
+   our $EVENT_INPUT   = 21;
    eval {
       require Net::SSLeay;
       Net::SSLeay->import( 1.30 );
@@ -30,19 +27,6 @@ BEGIN {
 
    no warnings 'redefine';
    my $old_new = \&POE::Wheel::ReadWrite::new;
-   my $old_set_filter = \&POE::Wheel::ReadWrite::set_filter;
-   my $old_set_input_filter = \&POE::Wheel::ReadWrite::set_input_filter;
-   my $old_set_output_filter = \&POE::Wheel::ReadWrite::set_output_filter;
-   my $old_rw_put = \&POE::Wheel::ReadWrite::put;
-   *POE::Wheel::ReadWrite::put = sub {
-      my $self = shift;
-      my $unique_id = $self->[POE::Wheel::ReadWrite::UNIQUE_ID()];
-      if (defined($self->[$EVENT_FLUSHED])) {
-         $self->[POE::Wheel::ReadWrite::EVENT_FLUSHED] = $self->[$EVENT_FLUSHED];
-         $self->[$EVENT_FLUSHED] = undef;
-      }
-      $old_rw_put->($self, @_);
-   };
    *POE::Wheel::ReadWrite::new = sub {
       my $class = shift;
       my %arg = @_;
@@ -71,75 +55,28 @@ BEGIN {
             }
          }
       );
-      $poe_kernel->state(
-         $self->[$PATCH] = ref($self) . "($unique_id) -> ssl patch",
-         sub {
-            my $type = $_[ARG0];
-            my $self = $_[ARG1];
-            if ($_[HEAP]->{self}->{PreFilter}) {
-               $_[HEAP]->{self}->{"PreFilter".ref($self).$self->[POE::Wheel::ReadWrite::UNIQUE_ID()]} = $_[HEAP]->{self}->{PreFilter}->clone()
-                  unless ($_[HEAP]->{self}->{"PreFilter".ref($self).$self->[POE::Wheel::ReadWrite::UNIQUE_ID()]});
-               if ($type eq "input") {
-                  $old_set_input_filter->($self, POE::Filter::Stackable->new(
-                     Filters => [
-                        $_[HEAP]->{self}->{"PreFilter".ref($self).$self->[POE::Wheel::ReadWrite::UNIQUE_ID()]},
-                        $self->get_input_filter()
-                     ]
-                  ));
-               } else {
-                  $old_set_output_filter->($self, POE::Filter::Stackable->new(
-                     Filters => [
-                        $_[HEAP]->{self}->{"PreFilter".ref($self).$self->[POE::Wheel::ReadWrite::UNIQUE_ID()]},
-                        $self->get_output_filter()
-                     ]
-                  ));
-               }
-            }
-         }
-      );
-      $poe_kernel->yield(ref($self) . "($unique_id) -> ssl patch" => "input" => $self);
-      $poe_kernel->yield(ref($self) . "($unique_id) -> ssl patch" => "output" => $self);
       return $self;
    };
+   my $old_rw_put = \&POE::Wheel::ReadWrite::put;
+   *POE::Wheel::ReadWrite::put = sub {
+      my $self = shift;
+      my $unique_id = $self->[POE::Wheel::ReadWrite::UNIQUE_ID()];
+      if (defined($self->[$EVENT_FLUSHED])) {
+         $self->[POE::Wheel::ReadWrite::EVENT_FLUSHED] = $self->[$EVENT_FLUSHED];
+         $self->[$EVENT_FLUSHED] = undef;
+      }
+      $old_rw_put->($self, @_);
+   };
    my $old_destroy = \&POE::Wheel::ReadWrite::DESTROY;
    *POE::Wheel::ReadWrite::DESTROY = sub {
       my $self = shift;
-      if ($self->[$PATCH]) {
-         $poe_kernel->state($self->[$PATCH]);
-         $self->[$PATCH] = undef;
-      }
       if ($self->[$HANDSHAKE]) {
          $poe_kernel->state($self->[$HANDSHAKE]);
          $self->[$HANDSHAKE] = undef;
       }
       return $old_destroy->($self, @_);
    };
-   *POE::Wheel::ReadWrite::set_filter = sub {
-      my $self = shift;
-      my $new_filter = shift;
-      my $unique_id = $self->[POE::Wheel::ReadWrite::UNIQUE_ID()];
-      my $ret = $old_set_filter->($self, $new_filter, @_);
-      $poe_kernel->yield(ref($self) . "($unique_id) -> ssl patch" => "input" => $self);
-      $poe_kernel->yield(ref($self) . "($unique_id) -> ssl patch" => "output" => $self);
-      return $ret;
-   };
-   *POE::Wheel::ReadWrite::set_input_filter = sub {
-      my $self = shift;
-      my $new_filter = shift;
-      my $unique_id = $self->[POE::Wheel::ReadWrite::UNIQUE_ID()];
-      my $ret = $old_set_input_filter->($self, $new_filter, @_);
-      $poe_kernel->yield(ref($self) . "($unique_id) -> ssl patch" => "input" => $self);
-      return $ret;
-   };
-   *POE::Wheel::ReadWrite::set_output_filter = sub {
-      my $self = shift;
-      my $new_filter = shift;
-      my $unique_id = $self->[POE::Wheel::ReadWrite::UNIQUE_ID()];
-      my $ret = $old_set_output_filter->($self, $new_filter, @_);
-      $poe_kernel->yield(ref($self) . "($unique_id) -> ssl patch" => "output" => $self);
-      return $ret;
-   };
-   #my $old_get_one  = \&POE::Filter::Stackable::get_one;
+   my $old_get_one  = \&POE::Filter::Stackable::get_one;
    *POE::Filter::Stackable::get_one = sub {
       my ($self) = @_;
       my $return = [ ];
@@ -191,9 +128,7 @@ BEGIN {
          $old_put->($self, $data, @_);
       }
    };
-   *POE::Filter::HTTPD::get_pending = sub {
-      return undef;
-   }
+   use warnings 'redefine';
 }
 
 require XSLoader;
@@ -209,90 +144,249 @@ sub checkForDoSendback {
    return 0;
 }
 
+sub PEMdataToX509 {
+   my $unblessed = shift;
+   my $x509 = shift;
+   my $bio = dataToBio($unblessed, $x509);
+   my $x509result = undef;
+   die "Error using x509: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+      unless ($x509result = Net::SSLeay::PEM_read_bio_X509($bio));
+   Net::SSLeay::BIO_free($bio);
+   return $x509result;
+}
+
+sub PEMdataToEVP_PKEY {
+   my $unblessed = shift;
+   my $crt = shift;
+   my $bio = dataToBio($unblessed, $crt);
+   my $evp_pkey = undef;
+   die "Error using cacrt: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+      unless ($evp_pkey = Net::SSLeay::PEM_read_bio_PrivateKey($bio));
+   Net::SSLeay::BIO_free($bio);
+   return $evp_pkey;
+}
+
+sub CTX_add_client_CA {
+   my $unblessed = shift;
+   my $ctx = shift;
+   my $x509 = shift;
+   my $err = Net::SSLeay::X509_STORE_add_cert(Net::SSLeay::CTX_get_cert_store($ctx), PEMdataToX509($unblessed, $x509));
+   die "Error using cacrt: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+      if ($err && ($err != 1));
+   $err = Net::SSLeay::CTX_add_client_CA($ctx, PEMdataToX509($unblessed, $x509));
+   die "Error using cacrt: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+      if ($err && ($err != 1));
+}
+
+sub dataToBio {
+   my $unblessed = shift;
+   my $data = shift;
+   my $bio = Net::SSLeay::BIO_new(Net::SSLeay::BIO_s_mem());
+   my $sent = Net::SSLeay::BIO_write($bio, $data);
+   print "Wrote ".$sent." of ".length($data)." bytes.\n"
+      if $unblessed->{debug};
+   die "Cannot write to bio!"
+      if (($sent) != length($data));
+   return $bio;
+}
+
 sub new {
    my $type = shift;
+
    my $params = {@_};
    my $self = bless({}, $type);
+
+   $self->{unblessed} = {};
    $self->{buffer} = '';
-   $self->{debug} = $params->{debug} || 0;
-   $self->{cacrl} = $params->{cacrl} || undef;
+   $self->{unblessed}->{debug} = $params->{debug} || 0;
+   $self->{cacrl} = $params->{cacrl}
+      if $self->{cacrl};
    $self->{client} = $params->{client} || 0;
    $self->{errorhandler} = $params->{errorhandler};
    $self->{params} = $params;
 
-   $self->{context} = Net::SSLeay::CTX_new();
+   $self->{context} =
+      ($params->{tls} || $params->{tls1_2}) ?
+                        ($params->{tls1_2}  ?
+         Net::SSLeay::CTX_tlsv1_2_new() :
+         Net::SSLeay::CTX_tlsv1_new()) :
+         Net::SSLeay::CTX_new();
 
-   if ((!$self->{client}) && (!$params->{"nohonor"})) { # Beim Apache: SSLHonorCipherOrder
-      Net::SSLeay::CTX_set_options($self->{context}, 0x00400000); # SSL_OP_CIPHER_SERVER_PREFERENCE
-   }
+   Net::SSLeay::CTX_set_options($self->{context}, 0x00400000) # SSL_OP_CIPHER_SERVER_PREFERENCE # Beim Apache: SSLHonorCipherOrder
+      if ((!$self->{client}) && (!$params->{"nohonor"}));
 
-   Net::SSLeay::CTX_use_RSAPrivateKey_file($self->{context}, $params->{key}, &Net::SSLeay::FILETYPE_PEM);
-   Net::SSLeay::CTX_use_certificate_file($self->{context}, $params->{crt}, &Net::SSLeay::FILETYPE_PEM);
+   my $err = undef;
    if ($params->{chain}) {
-      Net::SSLeay::CTX_use_certificate_chain_file($self->{context}, $params->{chain});
+      $err = Net::SSLeay::CTX_use_certificate_chain_file($self->{context}, $params->{chain});
+      die "Error using chain: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+         if ($err && ($err != 1));
    }
-   if ($params->{cacrt}) {
-      Net::SSLeay::CTX_load_verify_locations($self->{context}, $params->{cacrt}, '');
-      Net::SSLeay::CTX_set_client_CA_list($self->{context}, Net::SSLeay::load_client_CA_file($params->{cacrt}));
-      Net::SSLeay::CTX_set_verify_depth($self->{context}, 5);
+   if ($params->{keymem} || $params->{key}) {
+      if ($params->{keymem}) {
+         $err = Net::SSLeay::CTX_use_PrivateKey($self->{context}, PEMdataToEVP_PKEY($self->{unblessed}, $params->{keymem}));
+         print "Loaded keymem(".length($params->{keymem})." Bytes) with result ".$err."\n"
+            if $self->{unblessed}->{debug};
+      } else {
+         $err = Net::SSLeay::CTX_use_PrivateKey_file($self->{context}, $params->{key}, &Net::SSLeay::FILETYPE_PEM);
+         print "Loaded key from file ".$params->{key}." with result ".$err."\n"
+            if $self->{unblessed}->{debug};
+      }
+      die "Error using keymem: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+         if ($err && ($err != 1));
+   }
+   if ($params->{crtmem} || $params->{crt}) {
+      if ($params->{crtmem}) {
+         my $crt = PEMdataToX509($self->{unblessed}, $params->{crtmem});
+         $err = Net::SSLeay::CTX_use_certificate($self->{context}, $crt);
+         print "Loaded crtmem(".length($params->{crtmem})." Bytes/".$crt.") with result ".$err."\n"
+            if $self->{unblessed}->{debug};
+      } else {
+         # TODO:XXX:FIXME: Errorchecking!
+         $err = Net::SSLeay::CTX_use_certificate_file($self->{context}, $params->{crt}, &Net::SSLeay::FILETYPE_PEM);
+         print "Loaded crt from file ".$params->{crt}." with result ".$err."\n"
+            if $self->{unblessed}->{debug};
+      }
+      die "Error using crtmem: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+         if ($err && ($err != 1));
    }
 
-   if ($params->{cipher}) {
-      Net::SSLeay::CTX_set_cipher_list($self->{context}, $params->{cipher});
+   $err = undef;
+   if ($params->{cacrt}||
+       $params->{cacrtmem}) {
+      if ($params->{cacrtmem}) {
+         if (ref($params->{cacrtmem}) eq "ARRAY") {
+            foreach my $curcert (@{$params->{cacrtmem}}) {
+               $err = CTX_add_client_CA($self->{unblessed}, $self->{context}, $curcert);
+               last
+                  unless $err;
+            }
+         } else {
+            $err = CTX_add_client_CA($self->{unblessed}, $self->{context}, $params->{cacrtmem});
+            print "Loaded cacrtmem(".length($params->{cacrtmem})." Bytes) with result ".$err."\n"
+               if $self->{unblessed}->{debug};
+         }
+      } else {
+         $err = Net::SSLeay::CTX_load_verify_locations($self->{context}, $params->{cacrt}, '');
+         print "Loaded cacrt from file ".$params->{cacrt}." with result ".$err."\n"
+            if $self->{unblessed}->{debug};
+         $err = Net::SSLeay::CTX_set_client_CA_list($self->{context}, Net::SSLeay::load_client_CA_file($params->{cacrt}))
+            unless ($err && ($err == 1));
+         print "Set client cacrt from file ".$params->{cacrt}." with result ".$err."\n"
+            if $self->{unblessed}->{debug};
+      }
+      $err = Net::SSLeay::CTX_set_verify_depth($self->{context}, $params->{caverifydepth} || 5)
+         unless ($err && ($err == 1));
    }
+   die "Error using cacrt: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+      if ($err && ($err != 1));
 
+   $err = undef;
+   $err = Net::SSLeay::CTX_set_cipher_list($self->{context}, $params->{cipher})
+      if ($params->{cipher});
+   die "Error setting cipher: ".Net::SSLeay::ERR_error_string(ERR_get_error())
+      if ($err && ($err != 1));
+
+   $err = undef;
    $self->{rbio} = Net::SSLeay::BIO_new(Net::SSLeay::BIO_s_mem())
-      or die("Create rBIO_s_mem(): ".$!);
+      or die("Error creating r BIO: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error()));
    $self->{wbio} = Net::SSLeay::BIO_new(Net::SSLeay::BIO_s_mem())
-      or die("Create wBIO_s_mem(): ".$!);
+      or die("Error creating w BIO: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error()));
    $self->{ssl} = Net::SSLeay::new($self->{context});
-   Net::SSLeay::set_bio($self->{ssl}, $self->{rbio}, $self->{wbio});
+   $globalinfos->{int($self->{ssl})} = $self->{unblessed};
+   $err = Net::SSLeay::set_bio($self->{ssl}, $self->{rbio}, $self->{wbio});
+   die "Error setting r/w BIOs: ".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+      if ($err && ($err != 1));
 
    if ($params->{dhcert} ||
        $params->{dhcertmem}) {
       my $dhbio = undef;
       if ($params->{dhcertmem}) {
-         $dhbio = Net::SSLeay::BIO_new(Net::SSLeay::BIO_s_mem());
-         my $sent = Net::SSLeay::BIO_write($dhbio, $params->{dhcertmem});
-         die "Cannot write to dhcert bio!"
-            if (($sent) != length($params->{dhcertmem}));
+         $dhbio = dataToBio($self->{unblessed}, $params->{dhcertmem});
       } else {
          die "Cannot open dhcert file!"
-            unless (-f $params->{dhcert} && ($dhbio = Net::SSLeay::BIO_new_file($params->{dhcert}, "r")));
+            unless ((-s $params->{dhcert}) && ($dhbio = Net::SSLeay::BIO_new_file($params->{dhcert}, "r")));
       }
+      # TODO:XXX:FIXME: Errorchecking!
       my $dhret = Net::SSLeay::PEM_read_bio_DHparams($dhbio);
+      print "Loaded dhcert with result ".$err."\n"
+         if $self->{unblessed}->{debug};
       Net::SSLeay::BIO_free($dhbio);
       die "Couldn't set DH parameters!"
-         if (Net::SSLeay::set_tmp_dh($self->{ssl}, $dhret) < 0);
+         if (POE_FILTER_SSL_set_tmp_dh($self->{ssl}, $dhret) < 0);
+      print "Set dhcert params with result ".$err."\n"
+         if $self->{unblessed}->{debug};
+      #die "Couldn't set CTX DH parameters!"
+      #   if (POE_FILTER_SSL_CTX_set_tmp_dh($self->{context}, $dhret) < 0);
+      # TODO:XXX:FIXME: Errorchecking!
       my $rsa = Net::SSLeay::RSA_generate_key(2048, 73);
+      #die "Couldn't set RSA key!"
+      #   if (!Net::SSLeay::set_tmp_rsa($self->{ssl}, $rsa));
       die "Couldn't set RSA key!"
-         if (!Net::SSLeay::CTX_set_tmp_rsa($self->{context}, $rsa));
+         if (!POE_FILTER_SSL_CTX_set_tmp_rsa($self->{context}, $rsa));
+      print "Set dhrsa with result ".$err."\n"
+         if $self->{unblessed}->{debug};
       Net::SSLeay::RSA_free($rsa);
    }
-
-   if ($params->{clientcert}) {
-      my $orfilter = &Net::SSLeay::VERIFY_PEER
-                   | &Net::SSLeay::VERIFY_CLIENT_ONCE;
-      $orfilter |=  &Net::SSLeay::VERIFY_FAIL_IF_NO_PEER_CERT
-         if $params->{blockbadclientcert};
-      Net::SSLeay::set_verify($self->{ssl}, $orfilter, \&VERIFY);
+   my $orfilter = 0;
+   $orfilter |= &Net::SSLeay::VERIFY_PEER |
+                &Net::SSLeay::VERIFY_CLIENT_ONCE
+      if $params->{clientcert};
+   $orfilter |= &Net::SSLeay::VERIFY_FAIL_IF_NO_PEER_CERT
+      if $params->{clientcert} &&
+         $params->{blockbadclientcert};
+   # TODO:XXX:FIXME: Errorchecking!
+   #Net::SSLeay::CTX_set_verify($self->{context}, $orfilter, \&VERIFY);
+   Net::SSLeay::set_verify($self->{ssl}, $orfilter, \&VERIFY);
+   print "Set verify ".($params->{blockbadclientcert} ? "FORCE" : "")." ".$orfilter."\n"
+      if $self->{unblessed}->{debug};
+   if ($params->{sni}) {
+      my $err = Net::SSLeay::set_tlsext_host_name($self->{ssl}, $params->{sni});
+      print "Set sni with result ".$err."\n"
+         if $self->{unblessed}->{debug};
+      die "Error setting sni:".Net::SSLeay::ERR_error_string(Net::SSLeay::ERR_get_error())
+         if ($err && ($err != 1));
    }
-   
-   $globalinfos = [0, 0, []];
+   $self->{unblessed}->{ignoreVerifyErrors} = $params->{ignoreVerifyErrors}
+      if ($params->{ignoreVerifyErrors} &&
+     (ref($params->{ignoreVerifyErrors}) eq "ARRAY"));
 
    $self
 }
 
 sub VERIFY {
    my ($ok, $x509_store_ctx) = @_;
-   print "VERIFY!\n" if $debug;
-   $globalinfos->[0] = $ok ? 1 : 2 if ($globalinfos->[0] != 2);
-   $globalinfos->[1]++;
-   if (my $x = Net::SSLeay::X509_STORE_CTX_get_current_cert($x509_store_ctx)) {
-      push(@{$globalinfos->[2]},[Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($x)),
-                                 Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_issuer_name($x)),
-                                 X509_get_serialNumber($x)]);
+   my $ssl = Net::SSLeay::X509_STORE_CTX_get_ex_data($x509_store_ctx, POE_FILTER_SSL_get_ex_data_X509_STORE_CTX_idx());
+   my $unblessed = $globalinfos->{int($ssl)} || die;
+   print "VERIFY ".$ok
+      if $unblessed->{debug};
+   my $errcode = Net::SSLeay::X509_STORE_CTX_get_error($x509_store_ctx);
+   if ($unblessed->{ignoreVerifyErrors} &&
+  (ref($unblessed->{ignoreVerifyErrors}) eq "ARRAY") && (scalar(grep { $errcode == $_ }
+     @{$unblessed->{ignoreVerifyErrors}}))) {
+      $ok = 1;
+      print " -> ".$ok." (Ignoring error ".$errcode.")"
+         if $unblessed->{debug};
    }
+   print "\n"
+      if $unblessed->{debug};
+   $unblessed->{ok} = $ok ? 1 : 2
+      if (!defined($unblessed->{ok}) ||
+                  ($unblessed->{ok} != 2));
+   $unblessed->{chaincount}++;
+   # TODO:XXX:FIXME: Chainlength check
+   #X509_STORE_CTX_set_error($x509_store_ctx, X509_V_ERR_CERT_CHAIN_TOO_LONG)
+   #   if (X509_STORE_CTX_get_error_depth(ctx) > uuu);
+   # TODO:XXX:FIXME: No globalconfig
+   #    ssl = X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
+   #    mydata = SSL_get_ex_data(ssl, mydata_index);
+   #push(@{$unblessed->{ssls}}, int($x509_store_ctx));
+   if (my $x = Net::SSLeay::X509_STORE_CTX_get_current_cert($x509_store_ctx)) {
+      push(@{$unblessed->{chain}},[Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_subject_name($x)),
+                                   Net::SSLeay::X509_NAME_oneline(Net::SSLeay::X509_get_issuer_name($x)),
+                                   POE_FILTER_X509_get_serialNumber($x),
+                                   $errcode]);
+   }
+   Net::SSLeay::X509_STORE_CTX_set_error($x509_store_ctx, 0);
    return 1; # $ok; # 1=accept cert, 0=reject
 }
 
@@ -303,33 +397,41 @@ sub clone {
 
 sub get_one_start {
    my ($self, $data) = @_;
-   print "GETONESTART: NETWORK -> SSL -> POE: ".hexdump(join("", @$data))."\n" if $debug;
+   print "GETONESTART: NETWORK -> SSL -> POE: ".$self->hexdump(join("", @$data))."\n"
+      if $self->{unblessed}->{debug};
    $self->writeToSSLBIO(join("", @$data), $self->{accepted} ? 0 : 1);
    []
 }
 
 sub get_one {
    my $self = shift;
-   print "GETONE: BEGIN\n" if $debug;
+   print "GETONE: BEGIN\n"
+      if $self->{unblessed}->{debug};
    my @return = ();
    push(@return, $self) if ($self->doSSL() || $self->{buffer});
    my $data = Net::SSLeay::read($self->{ssl});
-   push(@return, $data) if $data;
-   print "GETONE: END: ".scalar(@return)."\n" if $debug;
+   push(@return, $data)
+      if $data;
+   print "GETONE: END: ".scalar(@return)."\n"
+      if $self->{unblessed}->{debug};
    [@return]
 }
 
 sub get {
-   print "GET: BEGIN\n" if $debug;
    my ($self, $chunks) = @_;
+   print "GET: BEGIN\n"
+      if $self->{unblessed}->{debug};
    my @return = ();
-   #print "GET:\n" if $debug;
+   #print "GET:\n"
+   #   if $self->{unblessed}->{debug};
    push(@return, $self) if ($self->doSSL() || $self->{buffer});
    foreach my $data (@$chunks) {
-      print "GET: NETWORK -> SSL -> POE: ".join("", @$data)."\n" if $debug;
+      print "GET: NETWORK -> SSL -> POE: ".join("", @$data)."\n"
+         if $self->{unblessed}->{debug};
       $self->writeToSSLBIO(join("", @$data));
       my $data = Net::SSLeay::read($self->{ssl});
-      print "GET: Read ".length($data)." bytes.\n" if $debug;
+      print "GET: Read ".length($data)." bytes.\n"
+         if $self->{unblessed}->{debug};
       push(@return, $data);
    }
    [@return]
@@ -337,7 +439,8 @@ sub get {
 
 sub put {
    my ($self, $chunks) = @_;
-   print "PUT: BEGIN\n" if $debug;
+   print "PUT: BEGIN\n"
+      if $self->{unblessed}->{debug};
    my @return = ();
    $self->doSSL();
    if ($self->{accepted}) {
@@ -350,14 +453,17 @@ sub put {
    }
    foreach my $data (@$chunks) {
       next if (ref($data) eq "POE::Filter::SSL");
-      print "PUT: POE -> SSL -> NETWORK: ".$data."\r\n" if $debug;
+      print "PUT: POE -> SSL -> NETWORK: ".$self->hexdump($data)."\r\n"
+         if $self->{unblessed}->{debug};
       if ($self->{accepted}) {
          $self->writeToSSL($data);
       } else {
-         push(@{$self->{sendbuf}}, $data) if ($data);
+         push(@{$self->{sendbuf}}, $data)
+            if ($data);
       }
    }
-   push(@return, $self->{buffer}) if $self->{buffer};
+   push(@return, $self->{buffer})
+      if $self->{buffer};
    $self->{buffer} = '';
    [@return]
 }
@@ -383,44 +489,6 @@ sub writeToSSLBIO {
    $self->doSSL() unless $nodoSSL;
 }
 
-#sub doHandshake {
-#   my $readWrite = shift;
-#   $readWrite = shift if (ref($readWrite) eq "POE::Filter::SSL");
-#   my @newFilters = @_;
-#   if (@newFilters) {
-#      if (ref($readWrite->get_input_filter()) eq "POE::Filter::SSL") {
-#         #print "ClientInput: ".$request."\n";
-#         if ($readWrite->get_input_filter()->handshakeDone(ignorebuf => 1)) {
-#            $readWrite->set_input_filter(POE::Filter::Stackable->new(
-#               Filters => [
-#                  $readWrite->get_input_filter(),
-#                  @newFilters
-#               ])
-#            );
-#         }
-#      }
-#   }
-#
-#   unless ((ref($readWrite->get_output_filter()) ne "POE::Filter::SSL") ||
-#               ($readWrite->get_output_filter()->handshakeDone())) {
-#      $readWrite->put();
-#      return 0;
-#   }
-#
-#   if (@newFilters) {
-#      if (ref($readWrite->get_output_filter()) eq "POE::Filter::SSL") {
-#         $readWrite->set_output_filter(POE::Filter::Stackable->new(
-#            Filters => [
-#               $readWrite->get_output_filter(),
-#               @newFilters
-#            ])
-#         );
-#      }
-#   }
-#   
-#   return 1;
-#}
-
 sub get_pending {
   return undef;
 }
@@ -428,14 +496,13 @@ sub get_pending {
 sub doSSL {
    my $self = shift;
    my $ret = 0;
-   print "SSLing..." if $debug;
+   print "SSLing..."
+      if $self->{unblessed}->{debug};
    unless ($self->{accepted}) {
       my $err = $self->{client} ?
          Net::SSLeay::connect($self->{ssl}) :
          Net::SSLeay::accept($self->{ssl});
       if ($err == 1) {
-         $self->{infos} = [((@$globalinfos)[0..2])];
-         $globalinfos = [0, 0, []];
          $self->{accepted}++;
          $ret++;
       } else {
@@ -478,7 +545,8 @@ sub doSSL {
    while (my $data = Net::SSLeay::BIO_read($self->{wbio})) {
       $self->{buffer} .= $data;
    }
-   print $ret."\n" if $debug;
+   print $ret."\n"
+      if $self->{unblessed}->{debug};
    return $ret;
 }
 
@@ -489,7 +557,7 @@ sub getCipher {
 
 sub clientCertExists {
    my $self = shift;
-   return ((ref($self->{infos}) eq "ARRAY") && ($self->{infos}->[1]));
+   return ((ref($self->{unblessed}->{chain}) eq "ARRAY") && ($self->{unblessed}->{chaincount}));
 }
 
 sub clientCertValid {
@@ -498,12 +566,12 @@ sub clientCertValid {
    if (defined($self->{cacrl})) {
       $valid = $self->clientCertNotOnCRL($self->{cacrl}) ? 1 : 0;
    }
-   return $self->clientCertExists() ? (($self->{infos}->[0] ne "2") && scalar(@{$self->{infos}->[2]}) && $valid) : undef;
+   return $self->clientCertExists() ? (($self->{unblessed}->{ok} ne "2") && scalar(@{$self->{unblessed}->{chain}}) && $valid) : undef;
 }
 
 sub clientCertIds {
    my $self = shift;
-   return $self->clientCertExists ? @{$self->{infos}->[2]} : undef;
+   return $self->clientCertExists ? @{$self->{unblessed}->{chain}} : undef;
 }
 
 sub clientCertNotOnCRL {
@@ -514,14 +582,15 @@ sub clientCertNotOnCRL {
       my $found = 0;
       my $badcrls = 0;
       my $jump = 0;
-      print("----- SSL Infos BEGIN ---------------"."\n") if $self->{debug};
-      foreach (@{$self->{infos}->[2]}) {
-         my $crlstatus = verify_serial_against_crl_file($crlfilename, $_->[2]);
+      print("----- SSL Infos BEGIN ---------------"."\n")
+         if $self->{unblessed}->{debug};
+      foreach (@{$self->{unblessed}->{chain}}) {
+         my $crlstatus = POE_FILTER_verify_serial_against_crl_file($crlfilename, $_->[2]);
          $badcrls++ if $crlstatus;
          $crlstatus = $crlstatus ? "INVALID (".($crlstatus !~ m,^CRL:, ? $self->hexdump($crlstatus) : $crlstatus).")" : "VALID";
          my $t = ("  " x $jump++);
          if (ref($_) eq "ARRAY") {
-            if ($self->{debug}){
+            if ($self->{unblessed}->{debug}){
                print(" ".$t."  |---[ Subcertificate ]---\n") if $t;
                print(" ".$t."  | Subject Name: ".$_->[0]."\n");
                print(" ".$t."  | Issuer Name : ".$_->[1]."\n");
@@ -529,11 +598,13 @@ sub clientCertNotOnCRL {
                print(" ".$t."  | CRL Status  : ".$crlstatus."\n");
             }
          } else {
-            print(" NOCERTINFOS!"."\n") if $self->{debug};
+            print(" NOCERTINFOS!"."\n")
+               if $self->{unblessed}->{debug};
             return 0;
          }
       }
-      print("----- SSL Infos END -----------------"."\n") if $self->{debug};
+      print("----- SSL Infos END -----------------"."\n")
+         if $self->{unblessed}->{debug};
       return 1 unless $badcrls;
    }
    return 0;
@@ -547,6 +618,8 @@ sub handshakeDone {
 
 sub DESTROY {
    my $self = shift;
+   #print "DESTROY: ".int($self->{ssl})."\n";
+   delete $globalinfos->{int($self->{ssl})};
    Net::SSLeay::free($self->{ssl})
       if $self->{ssl};
    Net::SSLeay::CTX_free($self->{context})
@@ -567,7 +640,7 @@ POE::Filter::SSL - The easiest and flexiblest way to SSL in POE!
 
 =head1 VERSION
 
-Version 0.31
+Version 0.41
 
 =head1 DESCRIPTION
 
@@ -676,6 +749,7 @@ By default I<POE::Filter::SSL> acts as a SSL server. To use it in client mode yo
 
 =item HTTPS-Server
 
+  use POE::Filter::SSL::PreFilter
   use POE::Filter::SSL;
   use POE::Component::Server::HTTP;
   use HTTP::Status;
@@ -715,7 +789,7 @@ By default I<POE::Filter::SSL> acts as a SSL server. To use it in client mode yo
 =item Advanced Example
 
 This example is an IMAP-Relay which forwards the connections to a IMAP server
-by username. It allows one the uncrypted transfer on port 143, with the option
+by username. It allows one the unencrypted transfer on port 143, with the option
 of SSL on the established connection (STARTTLS). On port 993 it allows one to do
 direct SSL.
 
@@ -974,25 +1048,35 @@ Returns a new I<POE::Filter::SSL> object. It accepts the following options:
 
 By default I<POE::Filter::SSL> acts as a SSL server. To use it in client mode, you have to set this option.
 
-=item crt
+=item crt{mem}
 
 The certificate file (.crt) for the server, a client certificate in client mode.
 
-=item key
+You are able to pass the already inmemory crt file as scalar via I<crtmem>.
+
+=item key{mem}
 
 The key file (.key) of the certificate (see I<crt> above).
 
-=item cacrt
+You are able to pass the already inmemory key file as scalar via I<keymem>.
 
-The ca certificate file (ca.crt), which is used to verificate the client certificates against a CA.
+=item cacrt{mem}
+
+The ca certificate file (ca.crt), which is used to verificate the client certificates against a CA. You can store multiple ca in one file, all of them gets imported.
+
+You are able to pass the already inmemory cacrt file as scalar via I<cacrtmem> or as an array ref of scalars, if you have multiple ca.
+
+=item caverifydepth
+
+By default the ca verify depth is 5, you can override this via this option.
 
 =item chain
 
-Chain certificate, you need it for example for startssl.org wich needs a intermedia certificates. Here you can configure it. You can generate this the following way:
+Chain certificate, you need it for example for startssl.org which needs a intermedia certificates. Here you can configure it. You can generate this the following way:
 
 cat client.crt intermediate.crt ca.crt > chain.pem
 
-In this case, you normaly have no I<key> and I<crt> option.
+In this case, you normalyly have no I<key> and I<crt> option. Currently it is not possible to pass this inmemory, only by file.
 
 =item cacrl
 
@@ -1010,7 +1094,19 @@ You are able to pass the already inmemory dhparam file as scalar(string) via I<d
 
 Only in server mode: Request during ssl handshake from the client a client certificat.
 
-B<WARNING:> If the client provides an untrusted or no client certficate, the connection is B<not> failing. You have to ask I<clientCertValid()> if the certicate is valid!
+B<WARNING:> If the client provides an untrusted or no client certificate, the connection is B<not> failing. You have to ask I<clientCertValid()> if the certificate is valid!
+
+=item sni
+
+Allows one to set the SNI hostname indication in first packet of handshake. See https://de.wikipedia.org/wiki/Server_Name_Indication
+
+=item tls
+
+Force in the handshake the use of tls, disables support for the obsolete SSL handshake.
+
+=item tls1_2
+
+Force in the handshake the use of tls in version 1.2, disables support for the obsolete SSL handshake.
 
 =item nohonor
 
@@ -1032,11 +1128,25 @@ Let OpenSSL deny the connection if there is no client certificate.
 
 B<WARNING:> If the client is listed in the CRL file or an invalid client certifiate has been sent, the connection will be established! You have to ask I<clientCertValid()> if you have the I<crl> option set on I<new()>, otherwise to ask I<clientCertNotOnCRL()> if the certificate is listed on your CRL file!
 
+=item ignoreVerifyErrors
+
+B<WARNING:> Before using this option, you should be realy sure that you know what you are doing!
+
+Specify to ignore specific errors on verifying the certificate chain: This is for example useful to be able to fetch the time from via secure and trusted TLS connection. In this case, your time is wrong, so must ignore time errors, which are 9: X509_V_ERR_CERT_NOT_YET_VALID (certificate is not yet valid) and 10: X509_V_ERR_CERT_HAS_EXPIRED (certificate has expired).
+
+The list of errors you can ignore can be found on the documentation:
+
+L<https://wiki.openssl.org/index.php/Manual:Verify(1)>
+
+Example:
+
+  ignoreVerifyErrors => [ 9, 10, ]
+
 =back
 
 =item handshakeDone(options)
 
-Returns I<true> if the handshake is done and all data for hanshake has been written out. It accepts the following options:
+Returns I<true> if the handshake is done and all data for handshake has been written out. It accepts the following options:
 
 =over 2
 
@@ -1090,7 +1200,7 @@ See the I<HTTPS-Server>, I<SSL on an established connection> and I<Client certif
 
 Returns I<true> if there is a client certificate, that might be untrusted.
 
-B<WARNING:> If the client provides an untrusted client certficate a client certicate that is listed in CRL, this function returns I<true>. You have to ask I<clientCertValid()> if the certicate is valid!
+B<WARNING:> If the client provides an untrusted client certificate a client certificate that is listed in CRL, this function returns I<true>. You have to ask I<clientCertValid()> if the certificate is valid!
 
 =item errorhandler
 
@@ -1104,7 +1214,7 @@ Do not report any error.
 
 =item I<CODE>
 
-Setting errorhandler to a reference of a function allows to be called it callback function with the following options:
+Setting errorhandler to a reference of a function allows one to be called it callback function with the following options:
 
 ARG1: POE:SSL::Filter instance
 
@@ -1112,7 +1222,7 @@ ARG2: Ref on a Hash with the following keys:
 
   ret        The return code of Net::SSLeay::connect (client) or Net::SSLeay::accept (server)
   ssl        The SSL context (SSL_CTX)
-  msg        The error message as text, as normaly reported via carp
+  msg        The error message as text, as normally reported via carp
   get_error  The error code of get_error the ssl context
   error      The error code of get_error without context
 
@@ -1126,7 +1236,7 @@ Do Carp/confess (stacktrace) on error.
 
 =item "carponetime"
 
-Report carp for one occurence only one time - over all!
+Report carp for one occurrence only one time - over all!
 
 =back
 
@@ -1151,7 +1261,13 @@ Example:
 
 =item VERIFY()
 
-=item X509_get_serialNumber()
+=item POE_FILTER_X509_get_serialNumber()
+
+=item POE_FILTER_SSL_CTX_set_tmp_dh()
+
+=item POE_FILTER_SSL_CTX_set_tmp_rsa()
+
+=item POE_FILTER_SSL_set_tmp_dh()
 
 =item clone()
 
@@ -1177,6 +1293,14 @@ Example:
 
 =item checkForDoSendback()
 
+=item CTX_add_client_CA()
+
+=item PEMdataToEVP_PKEY
+
+=item PEMdataToX509
+
+=item dataToBio
+
 =back
 
 =head1 AUTHOR
@@ -1219,13 +1343,13 @@ L<http://search.cpan.org/dist/POE-Filter-SSL>
 
 =head1 Commercial support
 
-Commercial support can be gained at <sslsupport at priv.de>.
+Commercial support can be gained at <sslsupport at cryptomagic.eu>.
 
 Used in our products, you can find on L<https://www.cryptomagic.eu/>
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2010-2016 Markus Schraeder, CryptoMagic GmbH, all rights reserved.
+Copyright 2010-2017 Markus Schraeder, CryptoMagic GmbH, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
